chiark / gitweb /
[PATCH] 038 release
[elogind.git] / namedev.c
index 52417e3793b655fcca041ff1ead58e7a0ba8d4ce..8a4436363d66bdcdff84a79a99e7412c671cf0f2 100644 (file)
--- a/namedev.c
+++ b/namedev.c
@@ -29,7 +29,6 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <errno.h>
-#include <time.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include <sys/sysinfo.h>
@@ -42,6 +41,7 @@
 #include "logging.h"
 #include "namedev.h"
 #include "klibc_fixups.h"
+#include "udevdb.h"
 
 static struct sysfs_attribute *find_sysfs_attribute(struct sysfs_class_device *class_dev, struct sysfs_device *sysfs_device, char *attr);
 
@@ -179,6 +179,37 @@ static int get_format_len(char **str)
        return -1;
 }
 
+/** Finds the lowest positive N such that <name>N isn't present in 
+ *  $(udevroot) either as a file or a symlink.
+ *
+ *  @param  name                Name to check for
+ *  @return                     0 if <name> didn't exist and N otherwise.
+ */
+static unsigned int find_free_number (struct udevice *udev, char *name)
+{
+       char temp[NAME_SIZE];
+       char path[NAME_SIZE];
+       struct udevice dev;
+       int result;
+
+       /* have to sweep the database for each lookup */
+       result = 0;
+       strncpy(temp, name, sizeof (temp));
+       while (1) {
+               if (udevdb_get_dev_byname(temp, path, &dev) != 0)
+                       goto found;
+               /* symlink might be stale if $(udevroot) isn't cleaned; check
+                * on major/minor to see if it's the same device
+                */
+               if (dev.major == udev->major && dev.minor == udev->minor)
+                       goto found;
+               snprintf (temp, sizeof(temp), "%s%d", name, ++result);
+       }
+
+found:
+       return result;
+}
+
 static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                         struct sysfs_class_device *class_dev,
                         struct sysfs_device *sysfs_device)
@@ -195,6 +226,7 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
        char *rest;
        int slen;
        struct sysfs_attribute *tmpattr;
+       unsigned int next_free_number;
 
        pos = string;
        while (1) {
@@ -274,6 +306,17 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                                        dbg("sysfa attribute '%s' not found", attr);
                                        break;
                                }
+                               /* strip trailing whitespace of matching value */
+                               if (isspace(tmpattr->value[strlen(tmpattr->value)-1])) {
+                                       i = len = strlen(tmpattr->value);
+                                       while (i > 0 &&  isspace(tmpattr->value[i-1]))
+                                               i--;
+                                       if (i < len) {
+                                               tmpattr->value[i] = '\0';
+                                               dbg("remove %i trailing whitespace chars from '%s'",
+                                                        len - i, tmpattr->value);
+                                       }
+                               }
                                strfieldcatmax(string, tmpattr->value, maxsize);
                                dbg("substitute sysfs value '%s'", tmpattr->value);
                        } else {
@@ -284,6 +327,13 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                        strfieldcatmax(string, "%", maxsize);
                        pos++;
                        break;
+               case 'e':
+                       next_free_number = find_free_number(udev, string);
+                       if (next_free_number > 0) {
+                               snprintf(temp2, sizeof(temp2), "%d", next_free_number);
+                               strfieldcatmax(string, temp2, maxsize);
+                       }
+                       break;
                default:
                        dbg("unknown substitution type '%%%c'", c);
                        break;
@@ -313,7 +363,6 @@ static struct bus_file {
        {}
 };
 
-#define SECONDS_TO_WAIT_FOR_FILE       10
 static void wait_for_device_to_initialize(struct sysfs_device *sysfs_device)
 {
        /* sleep until we see the file for this specific bus type show up this
@@ -327,14 +376,14 @@ static void wait_for_device_to_initialize(struct sysfs_device *sysfs_device)
        struct bus_file *b = &bus_files[0];
        struct sysfs_attribute *tmpattr;
        int found = 0;
-       int loop = SECONDS_TO_WAIT_FOR_FILE;
+       int loop = WAIT_FOR_FILE_SECONDS * WAIT_FOR_FILE_RETRY_FREQ;
 
        while (1) {
                if (b->bus == NULL) {
                        if (!found)
                                break;
-                       /* sleep to give the kernel a chance to create the file */
-                       sleep(1);
+                       /* give the kernel a chance to create the file */
+                       usleep(1000 * 1000 / WAIT_FOR_FILE_RETRY_FREQ);
                        --loop;
                        if (loop == 0)
                                break;
@@ -354,7 +403,8 @@ static void wait_for_device_to_initialize(struct sysfs_device *sysfs_device)
        }
        if (!found)
                dbg("did not find bus type '%s' on list of bus_id_files, "
-                   "contact greg@kroah.com", sysfs_device->bus);
+                   "please report to <linux-hotplug-devel@lists.sourceforge.net>",
+                   sysfs_device->bus);
 exit:
        return; /* here to prevent compiler warning... */
 }
@@ -372,7 +422,7 @@ static void fix_kernel_name(struct udevice *udev)
        }
 }
 
-static int execute_program(char *path, char *value, int len)
+static int execute_program(const char *path, char *value, int len)
 {
        int retval;
        int count;
@@ -381,28 +431,34 @@ static int execute_program(char *path, char *value, int len)
        pid_t pid;
        char *pos;
        char arg[PROGRAM_SIZE];
-       char *argv[sizeof(arg) / 2];
+       char *argv[(PROGRAM_SIZE / 2) + 1];
        int i;
 
+       strfieldcpy(arg, path);
        i = 0;
        if (strchr(path, ' ')) {
-               strfieldcpy(arg, path);
                pos = arg;
                while (pos != NULL) {
                        if (pos[0] == '\'') {
                                /* don't separate if in apostrophes */
                                pos++;
                                argv[i] = strsep(&pos, "\'");
-                               while (pos[0] == ' ')
+                               while (pos && pos[0] == ' ')
                                        pos++;
-               } else {
+                       } else {
                                argv[i] = strsep(&pos, " ");
                        }
                        dbg("arg[%i] '%s'", i, argv[i]);
                        i++;
                }
+               argv[i] =  NULL;
+               dbg("execute '%s' with parsed arguments", arg);
+       } else {
+               argv[0] = arg;
+               argv[1] = main_argv[1];
+               argv[2] = NULL;
+               dbg("execute '%s' with subsystem '%s' argument", arg, argv[1]);
        }
-       argv[i] =  NULL;
 
        retval = pipe(fds);
        if (retval != 0) {
@@ -414,17 +470,9 @@ static int execute_program(char *path, char *value, int len)
        switch(pid) {
        case 0:
                /* child */
-               close(STDOUT_FILENO);
-
-               /* dup write side of pipe to STDOUT */
-               dup(fds[1]);
-               if (argv[0] !=  NULL) {
-                       dbg("execute '%s' with given arguments", argv[0]);
-                       retval = execv(argv[0], argv);
-               } else {
-                       dbg("execute '%s' with main argument", path);
-                       retval = execv(path, main_argv);
-               }
+               /* dup2 write side of pipe to STDOUT */
+               dup2(fds[1], STDOUT_FILENO);
+               retval = execv(arg, argv);
 
                info(FIELD_PROGRAM " execution of '%s' failed", path);
                exit(1);
@@ -642,7 +690,6 @@ static struct sysfs_device *get_sysfs_device(struct sysfs_class_device *class_de
 {
        struct sysfs_device *sysfs_device;
        struct sysfs_class_device *class_dev_parent;
-       struct timespec tspec;
        int loop;
 
        /* Figure out where the device symlink is at.  For char devices this will
@@ -658,16 +705,14 @@ static struct sysfs_device *get_sysfs_device(struct sysfs_class_device *class_de
        if (class_dev_parent != NULL) 
                dbg("given class device has a parent, use this instead");
 
-       tspec.tv_sec = 0;
-       tspec.tv_nsec = 10000000;  /* sleep 10 millisec */
-       loop = 10;
+       loop = WAIT_FOR_FILE_SECONDS * WAIT_FOR_FILE_RETRY_FREQ;
        while (loop--) {
                if (udev_sleep) {
                        if (whitelist_search(class_dev)) {
                                sysfs_device = NULL;
                                goto exit;
                        }
-                       nanosleep(&tspec, NULL);
+                       usleep(1000 * 1000 / WAIT_FOR_FILE_RETRY_FREQ);
                }
 
                if (class_dev_parent)
@@ -689,11 +734,9 @@ device_found:
                if (sysfs_device->bus[0] != '\0')
                        goto bus_found;
 
-               loop = 10;
-               tspec.tv_nsec = 10000000;
                while (loop--) {
                        if (udev_sleep)
-                               nanosleep(&tspec, NULL);
+                               usleep(1000 * 1000 / WAIT_FOR_FILE_RETRY_FREQ);
                        sysfs_get_device_bus(sysfs_device);
                        
                        if (sysfs_device->bus[0] != '\0')
@@ -773,10 +816,12 @@ static int match_rule(struct config_device *dev, struct sysfs_class_device *clas
 
                /* execute external program */
                if (dev->program[0] != '\0') {
+                       char program[PROGRAM_SIZE];
+
                        dbg("check " FIELD_PROGRAM);
-                       apply_format(udev, dev->program, sizeof(dev->program),
-                                    class_dev, sysfs_device);
-                       if (execute_program(dev->program, udev->program_result, NAME_SIZE) != 0) {
+                       strfieldcpy(program, dev->program);
+                       apply_format(udev, program, sizeof(program), class_dev, sysfs_device);
+                       if (execute_program(program, udev->program_result, NAME_SIZE) != 0) {
                                dbg(FIELD_PROGRAM " returned nonzero");
                                goto try_parent;
                        } else {